package at.tuwien.endpoints;

import at.tuwien.api.auth.LoginRequestDto;
import at.tuwien.api.auth.RefreshTokenRequestDto;
import at.tuwien.api.auth.SignupRequestDto;
import at.tuwien.api.error.ApiErrorDto;
import at.tuwien.api.keycloak.TokenDto;
import at.tuwien.api.user.*;
import at.tuwien.entities.database.Database;
import at.tuwien.entities.user.User;
import at.tuwien.exception.*;
import at.tuwien.mapper.UserMapper;
import at.tuwien.service.AuthenticationService;
import at.tuwien.service.DatabaseService;
import at.tuwien.service.UserService;
import at.tuwien.utils.UserUtil;
import io.micrometer.observation.annotation.Observed;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.media.ArraySchema;
import io.swagger.v3.oas.annotations.media.Content;
import io.swagger.v3.oas.annotations.media.Schema;
import io.swagger.v3.oas.annotations.responses.ApiResponse;
import io.swagger.v3.oas.annotations.responses.ApiResponses;
import io.swagger.v3.oas.annotations.security.SecurityRequirement;
import jakarta.servlet.http.HttpServletRequest;
import jakarta.validation.Valid;
import jakarta.validation.constraints.NotNull;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.bind.annotation.*;

import java.security.Principal;
import java.util.List;
import java.util.UUID;

@Log4j2
@CrossOrigin(origins = "*")
@RestController
@RequestMapping(path = "/api/user")
public class UserEndpoint {

    private final UserMapper userMapper;
    private final UserService userService;
    private final DatabaseService databaseService;
    private final AuthenticationService authenticationService;

    @Autowired
    public UserEndpoint(UserMapper userMapper, UserService userService, DatabaseService databaseService,
                        AuthenticationService authenticationService) {
        this.userMapper = userMapper;
        this.userService = userService;
        this.databaseService = databaseService;
        this.authenticationService = authenticationService;
    }

    @GetMapping
    @Transactional(readOnly = true)
    @Observed(name = "dbrepo_metadata_users_list")
    @Operation(summary = "Find all users")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "200",
                    description = "List users",
                    content = {@Content(
                            mediaType = "application/json",
                            array = @ArraySchema(schema = @Schema(implementation = UserBriefDto.class)))}),
    })
    public ResponseEntity<List<UserBriefDto>> findAll() {
        log.debug("endpoint find all users");
        final List<UserBriefDto> users = userService.findAll()
                .stream()
                .map(userMapper::userToUserBriefDto)
                .toList();
        log.trace("find all users resulted in users {}", users);
        return ResponseEntity.ok(users);
    }

    @PostMapping
    @Transactional(rollbackFor = {ServiceException.class, ServiceConnectionException.class})
    @PreAuthorize("!isAuthenticated()")
    @Observed(name = "dbrepo_metadata_user_create")
    @Operation(summary = "Create user")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "201",
                    description = "Created user",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = UserBriefDto.class))}),
            @ApiResponse(responseCode = "400",
                    description = "Parameters are not well-formed (likely email)",
                    content = {@Content(mediaType = "application/json")}),
            @ApiResponse(responseCode = "404",
                    description = "default role not found",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
            @ApiResponse(responseCode = "409",
                    description = "User with username already exists",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
            @ApiResponse(responseCode = "417",
                    description = "User with e-mail already exists",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
    })
    public ResponseEntity<UserBriefDto> create(@NotNull @Valid @RequestBody SignupRequestDto data)
            throws UserExistsException, EmailExistsException, ServiceException, ServiceConnectionException,
            UserNotFoundException {
        log.debug("endpoint create a user, data.username={}", data.getUsername());
        userService.validateUsernameNotExists(data.getUsername());
        userService.validateEmailNotExists(data.getEmail());
        authenticationService.create(data);
        final at.tuwien.api.keycloak.UserDto keycloakUserDto = authenticationService.findByUsername(data.getUsername());
        final User user = userService.create(data, keycloakUserDto.getId());
        final UserBriefDto dto = userMapper.userToUserBriefDto(user);
        log.trace("create user resulted in dto {}", dto);
        return ResponseEntity.status(HttpStatus.CREATED)
                .body(dto);
    }

    @PostMapping("/token")
    @Observed(name = "dbrepo_metadata_user_token")
    @Operation(summary = "Obtain user token")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "202",
                    description = "Obtained user token",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = TokenDto.class))}),
    })
    public ResponseEntity<TokenDto> getToken(@NotNull @Valid @RequestBody LoginRequestDto data)
            throws ServiceException, ServiceConnectionException, UserNotFoundException, CredentialsInvalidException,
            AccountNotSetupException {
        log.debug("endpoint get token, data.username={}", data.getUsername());
        /* check */
        final TokenDto token = authenticationService.obtainToken(data);
        try {
            userService.findByUsername(data.getUsername());
        } catch (UserNotFoundException e) {
            /* need to sync */
            log.debug("User with username {} does not exist in metadata database yet", data.getUsername());
            final SignupRequestDto request = SignupRequestDto.builder()
                    .username(data.getUsername())
                    .email("noreply@example.com")
                    .password(data.getPassword())
                    .build();
            userService.create(request, authenticationService.findByUsername(data.getUsername()).getId());
            log.info("Fetched user information from auth service and stored it into metadata database");
        }
        return ResponseEntity.accepted()
                .body(token);
    }

    @PutMapping("/token")
    @Observed(name = "dbrepo_metadata_user_refresh_token")
    @Operation(summary = "Refresh user token")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "202",
                    description = "Refreshed user token",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = TokenDto.class))}),
    })
    public ResponseEntity<TokenDto> refreshToken(@NotNull @Valid @RequestBody RefreshTokenRequestDto data)
            throws ServiceConnectionException, CredentialsInvalidException {
        log.debug("endpoint refresh token");
        /* check */
        final TokenDto token = authenticationService.refreshToken(data.getRefreshToken());
        return ResponseEntity.accepted()
                .body(token);
    }

    @GetMapping("/{userId}")
    @Transactional(readOnly = true)
    @PreAuthorize("isAuthenticated()")
    @Observed(name = "dbrepo_metadata_user_find")
    @Operation(summary = "Get a user info", security = {@SecurityRequirement(name = "bearerAuth"), @SecurityRequirement(name = "basicAuth")})
    @ApiResponses(value = {
            @ApiResponse(responseCode = "200",
                    description = "Found user",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = UserDto.class))}),
            @ApiResponse(responseCode = "403",
                    description = "Find user is not permitted",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
            @ApiResponse(responseCode = "404",
                    description = "User was not found",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
    })
    public ResponseEntity<UserDto> find(@NotNull @PathVariable("userId") UUID userId,
                                        @NotNull Principal principal) throws NotAllowedException,
            UserNotFoundException {
        log.debug("endpoint find a user, userId={}", userId);
        /* check */
        final User user = userService.findById(userId);
        if (!user.equals(principal)) {
            if (!UserUtil.hasRole(principal, "admin")) {
                log.error("Failed to find user: foreign user");
                throw new NotAllowedException("Failed to find user: foreign user");
            }
        }
        final UserDto dto = userMapper.userToUserDto(user);
        return ResponseEntity.ok()
                .body(dto);
    }

    @PutMapping("/{userId}")
    @Transactional
    @PreAuthorize("hasAuthority('modify-user-information')")
    @Observed(name = "dbrepo_metadata_user_modify")
    @Operation(summary = "Modify user information", security = {@SecurityRequirement(name = "bearerAuth"), @SecurityRequirement(name = "basicAuth")})
    @ApiResponses(value = {
            @ApiResponse(responseCode = "202",
                    description = "Modified user information",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = UserDto.class))}),
            @ApiResponse(responseCode = "400",
                    description = "Modify user query is malformed",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
    })
    public ResponseEntity<UserDto> modify(@NotNull @PathVariable("userId") UUID userId,
                                          @NotNull @Valid @RequestBody UserUpdateDto data,
                                          @NotNull Principal principal) throws ServiceException,
            ServiceConnectionException, NotAllowedException, UserNotFoundException, DatabaseNotFoundException {
        log.debug("endpoint modify a user, userId={}, data={}", userId, data);
        User user = userService.findById(userId);
        if (!user.equals(principal)) {
            log.error("Failed to modify user: not current user");
            throw new NotAllowedException("Failed to modify user: not current user");
        }
        user = userService.modify(user, data);
        final UserDto dto = userMapper.userToUserDto(user);
        return ResponseEntity.accepted()
                .body(dto);
    }

    @PutMapping("/{userId}/password")
    @Transactional
    @PreAuthorize("isAuthenticated()")
    @Observed(name = "dbrepo_metadata_user_password_modify")
    @Operation(summary = "Modify user password", security = {@SecurityRequirement(name = "bearerAuth"), @SecurityRequirement(name = "basicAuth")})
    @ApiResponses(value = {
            @ApiResponse(responseCode = "202",
                    description = "Modified user password",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = UserDto.class))}),
    })
    public ResponseEntity<?> password(@NotNull @PathVariable("userId") UUID userId,
                                      @NotNull @Valid @RequestBody UserPasswordDto data,
                                      @NotNull Principal principal) throws NotAllowedException, ServiceException,
            ServiceConnectionException, UserNotFoundException, DatabaseNotFoundException {
        log.debug("endpoint modify a user password, userId={}, data.password=(hidden)", userId);
        User user = userService.findById(userId);
        if (!user.equals(principal)) {
            log.error("Failed to modify user password: not current user");
            throw new NotAllowedException("Failed to modify user password: not current user");
        }
        user = userService.findByUsername(principal.getName());
        userService.updatePassword(user, data);
        authenticationService.updatePassword(user, data);
        for (Database database : databaseService.findAllAccess(userId)) {
            databaseService.updatePassword(database, user);
        }
        return ResponseEntity.accepted()
                .build();
    }

}
