package at.tuwien.service.impl;

import at.tuwien.api.database.ViewCreateDto;
import at.tuwien.entities.database.Database;
import at.tuwien.entities.database.View;
import at.tuwien.entities.user.User;
import at.tuwien.exception.*;
import at.tuwien.gateway.DataServiceGateway;
import at.tuwien.gateway.SearchServiceGateway;
import at.tuwien.mapper.QueryMapper;
import at.tuwien.mapper.ViewMapper;
import at.tuwien.repository.DatabaseRepository;
import at.tuwien.service.ViewService;
import com.google.common.hash.Hashing;
import lombok.extern.log4j.Log4j2;
import net.sf.jsqlparser.JSQLParserException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.nio.charset.StandardCharsets;
import java.util.LinkedList;
import java.util.List;
import java.util.Optional;

@Log4j2
@Service
public class ViewServiceImpl implements ViewService {

    private final ViewMapper viewMapper;
    private final QueryMapper queryMapper;
    private final DataServiceGateway dataServiceGateway;
    private final DatabaseRepository databaseRepository;
    private final SearchServiceGateway searchServiceGateway;

    @Autowired
    public ViewServiceImpl(ViewMapper viewMapper, QueryMapper queryMapper, DataServiceGateway dataServiceGateway,
                           DatabaseRepository databaseRepository, SearchServiceGateway searchServiceGateway) {
        this.viewMapper = viewMapper;
        this.queryMapper = queryMapper;
        this.dataServiceGateway = dataServiceGateway;
        this.databaseRepository = databaseRepository;
        this.searchServiceGateway = searchServiceGateway;
    }

    @Override
    public View findById(Database database, Long viewId) throws ViewNotFoundException {
        final Optional<View> optional = database.getViews()
                .stream()
                .filter(v -> v.getId().equals(viewId))
                .findFirst();
        if (optional.isEmpty()) {
            log.error("Failed to find view with id {}", viewId);
            throw new ViewNotFoundException("Failed to find view with id " + viewId);
        }
        return optional.get();
    }

    @Override
    @Transactional(readOnly = true)
    public List<View> findAll(Database database, User user) {
        if (user == null) {
            return database.getViews()
                    .stream()
                    .filter(View::getIsPublic)
                    .toList();
        }
        return database.getViews()
                .stream()
                .filter(v -> v.getIsPublic() || v.getCreatedBy().equals(user.getId()))
                .toList();
    }

    @Override
    @Transactional
    public void delete(View view) throws ServiceException, ServiceConnectionException, DatabaseNotFoundException,
            ViewNotFoundException, SearchServiceException, SearchServiceConnectionException {
        /* delete in data service */
        dataServiceGateway.deleteView(view.getDatabase().getId(), view.getId());
        /* delete in metadata database */
        view.getDatabase().getViews().remove(view);
        final Database database = databaseRepository.save(view.getDatabase());
        /* update in search service */
        searchServiceGateway.update(database);
        log.info("Deleted view with id {}", view.getId());
    }

    @Override
    @Transactional
    public View create(Database database, User creator, ViewCreateDto data) throws MalformedException, ServiceException,
            ServiceConnectionException, DatabaseNotFoundException, SearchServiceException, SearchServiceConnectionException {
        /* create in metadata database */
        final View view = View.builder()
                .vdbid(database.getId())
                .database(database)
                .name(data.getName())
                .internalName(viewMapper.nameToInternalName(data.getName()))
                .createdBy(creator.getId())
                .creator(creator)
                .identifiers(new LinkedList<>())
                .query(data.getQuery())
                .queryHash(Hashing.sha256()
                        .hashString(data.getQuery(), StandardCharsets.UTF_8)
                        .toString())
                .isInitialView(false)
                .isPublic(data.getIsPublic())
                .build();
        /* create in data service */
        data.setName(view.getInternalName());
        dataServiceGateway.createView(database.getId(), data);
        try {
            view.setColumns(viewMapper.tableColumnsToViewColumns(view, queryMapper.parseColumns(data.getQuery(), database)));
        } catch (JSQLParserException e) {
            throw new MalformedException("Failed to parse columns from view: " + e.getMessage(), e);
        }
        database.getViews()
                .add(view);
        database = databaseRepository.save(database);
        final Optional<View> optional = database.getViews()
                .stream()
                .filter(v -> v.getInternalName().equals(view.getInternalName()))
                .findFirst();
        if (optional.isEmpty()) {
            log.error("Failed to find created view");
            throw new MalformedException("Failed to find created view");
        }
        /* update in search service */
        searchServiceGateway.update(database);
        log.info("Created view with id {}", optional.get().getId());
        return optional.get();
    }

}
